#define PROGNAME  "lmpc2pbm"
#define VERSION   "1.0"
#define COPYRIGHT "Copyright (C) 2008  J.A.Bezemer@opensourcepartners.nl  GPL>=2"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXL  100000		/* size of matrix allocated: lines... */
#define MAXW  24		/* ...and width in bytes (= 8 pixels) */


typedef enum
{
  STATE_INIT,
  STATE_IDLE,
  STATE_1BSEEN,
  STATE_1B43SEEN,
  STATE_1B44SEEN,
  STATE_ENTERS,
  STATE_1B44XXSEEN,
  STATE_COPYING,
  STATE_1B42SEEN,
} state_t;


int
main (int argc, char **argv)
{
  state_t state = STATE_INIT;
  int papersize = -1;
  int inktype = -1;
  int margin = -1;
  int bytes2copy = -1;
  int i;
  int line = -1;		/* we do ++ before printing first line */
  int pos = -1;
  int maxpos = -1;
  unsigned char matrix[MAXL][MAXW];

  /* Force width: cutlines are always printed in 16bytes; and 24mm uses full
     16bytes too, so print head probably is 16*8=128 pixels. <- LMPC2.
     LMPC"1": Manual suggests 12*8=96 pixels. */
  int forcew = 16;


  if (argc == 2)
    {
      forcew = atoi (argv[1]);
      if (forcew <= -2 || forcew > MAXW)
	forcew = 0;
    }

  if (argc > 2 || forcew == 0)
    {
      fprintf (stderr,
	       "\n"
	       "USAGE: %s [forcewidth]  < printfile  > pbmfile\n"
	       "\n"
	       "Converts Dymo LabelManager PC (PC-II) printfile to PBM image format.\n"
	       "Currently compiled to handle %d lines of %d*8 pixels.\n"
	       "\n"
	       "forcewidth:  pbm width in bytes (=8 pixel multiples)  -1 for minimum width;\n"
	       "             12 for LMPC; 16 for LMPC2 (default)\n"
	       "\n" "Version %s  %s\n" "\n", PROGNAME, MAXL, MAXW, VERSION,
	       COPYRIGHT);
      exit (1);
    }


  memset (matrix, 0, MAXL * MAXW);

  while (!feof (stdin))
    {
      i = fgetc (stdin);
      if (i == EOF)
	break;

      switch (state)
	{
	case STATE_INIT:
	  papersize = i;
	  state = STATE_IDLE;
	  break;


	case STATE_IDLE:
	  if (i == 0x1b)
	    state = STATE_1BSEEN;
	  else
	    {
	      fprintf (stderr, "State %d unknown char 0x%02x, aborting\n",
		       state, i);
	      exit (1);
	    }
	  break;


	case STATE_1BSEEN:
	  if (i == 0x42)
	    state = STATE_1B42SEEN;
	  else if (i == 0x43)
	    state = STATE_1B43SEEN;
	  else if (i == 0x44)
	    state = STATE_1B44SEEN;
	  else
	    {
	      fprintf (stderr, "State %d unknown char 0x%02x, aborting\n",
		       state, i);
	      exit (1);
	    }
	  break;


	case STATE_1B43SEEN:
	  inktype = i;
	  state = STATE_IDLE;
	  break;


	case STATE_1B44SEEN:
	  bytes2copy = i;
	  if (bytes2copy > 0)
	    state = STATE_1B44XXSEEN;
	  else
	    state = STATE_ENTERS;
	  break;


	case STATE_ENTERS:
	  if (i == 0x16)
	    {
	      line++;
	      if (line >= MAXL)
		{
		  fprintf (stderr,
			   "Exceeded max lines (%d, recompile to enlarge), aborting\n",
			   MAXL);
		  exit (1);
		}
	    }

	  else if (i == 0x1b)
	    state = STATE_1BSEEN;

	  else
	    {
	      fprintf (stderr, "State %d unknown char 0x%02x, aborting\n",
		       state, i);
	      exit (1);
	    }
	  break;


	case STATE_1B44XXSEEN:
	  if (i == 0x16)
	    {
	      line++;		/* checked below */
	      pos = margin;

	      if (margin < 0)
		{
		  fprintf (stderr,
			   "Printing with margin %d (missing 1b-42-mm), aborting\n",
			   margin);
		  exit (1);
		}

	      if (bytes2copy <= 0)
		{
		  fprintf (stderr, "Printing with bytes2copy %d, aborting\n",
			   bytes2copy);
		  exit (1);
		}

	      if (margin + bytes2copy > maxpos)
		maxpos = margin + bytes2copy;

	      if (maxpos >= MAXW)
		{
		  fprintf (stderr,
			   "Exceeded max width (%d, recompile to enlarge), aborting\n",
			   MAXW);
		  exit (1);
		}

	      if (line >= MAXL)
		{
		  fprintf (stderr,
			   "Exceeded max lines (%d, recompile to enlarge), aborting\n",
			   MAXL);
		  exit (1);
		}

	      state = STATE_COPYING;
	    }
	  else
	    {
	      fprintf (stderr, "State %d unknown char 0x%02x, aborting\n",
		       state, i);
	      exit (1);
	    }
	  break;


	case STATE_COPYING:
	  /* printf ("line %2d pos %2d byte 0x%02x\n", line, pos, i); */
	  matrix[line][pos] = (unsigned char) i;
	  pos++;
	  bytes2copy--;
	  if (bytes2copy <= 0)
	    state = STATE_IDLE;
	  break;


	case STATE_1B42SEEN:
	  margin = i;
	  state = STATE_IDLE;
	  break;


	default:
	  fprintf (stderr, "Unknown state %d, aborting\n", state);
	  exit (1);
	}
    }


#if 0
  printf ("Printed %d lines, %d pos\n", line + 1, maxpos);
  for (i = 0; i <= line; i++)
    {
      for (pos = 0; pos < maxpos; pos++)
	printf ("0x%02x ", matrix[i][pos]);
      printf ("\n");
    }
#endif

  if (forcew > 0)
    {
      if (maxpos > forcew)
	fprintf (stderr, "Warning: maxpos=%d truncated to forcew=%d.\n",
		 maxpos, forcew);
      maxpos = forcew;
    }

  printf ("P4\n%d %d\n", maxpos * 8, line + 1);
  for (i = 0; i <= line; i++)	/* line==0 means 1 line printed */
    {
      fwrite (&(matrix[i][0]), 1, maxpos, stdout);
    }

  return 0;
}
