#define PROGNAME  "pbm2lmpc"
#define VERSION   "1.01"
#define COPYRIGHT "Copyright (C) 2008  J.A.Bezemer@opensourcepartners.nl  GPL>=2"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


#define BUFLEN  4000


void
usage (int err)
{
  fprintf (err ? stderr : stdout,
	   "\n"
	   "USAGE: %s [options]  < inputpbm  > printfile\n"
	   "\n"
	   "Convert binary PBM files to Dymo LabelManager PC and PC-II print format.\n"
	   "PBM should be vertical with correct width (unless -c), also see pnmflip.\n"
	   "To adjust margins, see dypbmfixmargins.\n"
	   "\n"
	   "Options:\n"
	   "  -c             Allow input wider than labelwidth and crop it\n"
	   "  -E             Read input until EOF regardless of PBM header (use if you\n"
	   "                 don't know the length in advance; still write >0 in header)\n"
	   "  -h             This help\n"
	   "  -i<inktype>    Ink/tape type as number or just \"black-on-white\" etc.;\n"
	   "                       0  black on white/clear\n"
	   "                       1  black on blue\n"
	   "                       2  black on red\n"
	   "                       3  black on silver\n"
	   "                       4  black on yellow\n"
	   "                       5  black on gold\n"
	   "                       6  black on green\n"
	   "                       7  black on fluorescent green  (\"fluogreen\")\n"
	   "                       8  black on fluorescent red    (\"fluored\")\n"
	   "                       9  white on clear\n"
	   "                      10  white on black\n"
	   "                      11  blue on white/clear\n"
	   "                      12  red on white/clear\n"
	   "  -l<bytes>      Labelwidth in bytes: 12 for LMPC (96 pixels), 16 for LMPC2\n"
	   "                 (128 pixels; default)\n"
	   "  -p<papersize>  Papersize: 0=6mm 1=9mm 2=12mm(default) 3=18/19mm 4=24mm\n"
	   "                 or just \"6mm\" up to \"24mm\" - \"1/4in\" up to \"1in\"\n"
	   "                 (Note that narrow tapes are centered on the print head; wide\n"
	   "                 tapes can't be printed fully: only 96 or 128 pixels at 180dpi)\n"
	   "  -t             Passthrough only for print files (matching ** 1B 43 ** 1B)\n"
	   "  -T             Passthrough for all non-PBM files\n"
	   "\n" "Version %s  %s\n" "\n", PROGNAME, VERSION, COPYRIGHT);
  exit (err);
}


char *
fgets0len (char *buf, int bufsize, FILE * stream, int *buffilled)
/* Read stream until newline OR zero-byte OR EOF; at most bufsize-1 chars.
   Always put extra zero-byte after data.
   Return buf when anything read; NULL on immediate error/EOF (nothing read).
   Return *buffilled with meaningful #bytes in buffer, _not_ counting
   the added zero-byte.
*/
{
  unsigned char *bufu = (void *) buf;
  int i;

  *buffilled = 0;

  while (*buffilled < bufsize - 1)
    {
      i = fgetc (stream);

      if (i == EOF)
	{
	  if (*buffilled == 0)
	    return NULL;
	  else
	    break;		/* finalize and return */
	}

      *bufu = (unsigned char) i;
      bufu++;
      (*buffilled)++;

      if (i == '\n' || i == '\0')
	break;			/* finalize and return */
    }
  /* normally: buf full */

  *bufu = '\0';
  return buf;
}


#if 0
int
fgetstest ()
{
  char buf[1000];
  int i;
  char *cp;
  FILE *f;


  /* fgets() */

  f = fopen ("/tmp/testfgets", "ro");

  for (i = 0; i < 1000; i++)
    ((unsigned char *) buf)[i] = (i + 1) & 0xff;

  cp = fgets (buf, 7, f);
  printf ("fgets: cp=0x%p\n", cp);

  printf (">%.30s<", buf);
  for (i = 0; i < 10; i++)
    printf (" %02x", ((unsigned char *) buf)[i]);
  printf ("\n");

  fread (buf, 1, 1, f);
  printf ("next fread: %02x\n", ((unsigned char *) buf)[0]);


  /* fgets0len() */

  rewind (f);

  for (i = 0; i < 1000; i++)
    ((unsigned char *) buf)[i] = (i + 1) & 0xff;

  cp = fgets0len (buf, 7, f, &i);
  printf ("fgets0len: cp=0x%p i=%d\n", cp, i);

  printf (">%.30s<", buf);
  for (i = 0; i < 10; i++)
    printf (" %02x", ((unsigned char *) buf)[i]);
  printf ("\n");

  fread (buf, 1, 1, f);
  printf ("next fread: %02x\n", ((unsigned char *) buf)[0]);


  return 0;
}
#endif /* 0 */


int
main (int argc, char **argv)
{
  int labelwb = 16;		/* label width in bytes */
  int cropwidth = 0;		/* allow cropping */
  int papersize = 0x02;
  int inktype = 0x00;
  int passthrough = 0;
  int line = 0;			/* output line counter; 0 = follow pbm header
				   and ignore trailing garbage; or
				   -1 = until EOF (ignore header) */

  char buf[BUFLEN + 5];
  unsigned char *bufu = (void *) buf;
  char *cp;
  int p4seen = 0;
  int headerwidth = -1;
  int headerlines = -1;
  int filewb = -1;		/* file width in bytes */

  int i;
  int currmargin = 0;
  int doingempty = 0;		/* >0: counter up to maxempty */
  int maxempty = 32;		/* possibly infinite */
  int leftzero, rightzero;


  /* return fgetstest (); */


  /* check options */

  while (1)
    {
      i = getopt (argc, argv, "cEhi:l:p:tT");

      if (i < 0)
	{
	  if (optind < argc)
	    usage (1);
	  break;
	}

      switch (i)
	{
	  /* Keep alphabetical */

	case 'c':
	  cropwidth = 1;
	  break;

	case 'E':
	  line = -1;
	  break;

	case 'h':
	  usage (0);

	case 'i':
	  if (!strcasecmp (optarg, "black on white") ||
	      !strcasecmp (optarg, "black-on-white") ||
	      !strcasecmp (optarg, "blackonwhite") ||
	      !strcasecmp (optarg, "black-white") ||
	      !strcasecmp (optarg, "blackwhite") ||
	      !strcasecmp (optarg, "black on clear") ||
	      !strcasecmp (optarg, "black-on-clear") ||
	      !strcasecmp (optarg, "blackonclear") ||
	      !strcasecmp (optarg, "black-clear") ||
	      !strcasecmp (optarg, "blackclear"))
	    inktype = 0x00;
	  else if (!strcasecmp (optarg, "black on blue") ||
		   !strcasecmp (optarg, "black-on-blue") ||
		   !strcasecmp (optarg, "blackonblue") ||
		   !strcasecmp (optarg, "black-blue") ||
		   !strcasecmp (optarg, "blackblue"))
	    inktype = 0x01;
	  else if (!strcasecmp (optarg, "black on red") ||
		   !strcasecmp (optarg, "black-on-red") ||
		   !strcasecmp (optarg, "blackonred") ||
		   !strcasecmp (optarg, "black-red") ||
		   !strcasecmp (optarg, "blackred"))
	    inktype = 0x02;
	  else if (!strcasecmp (optarg, "black on silver") ||
		   !strcasecmp (optarg, "black-on-silver") ||
		   !strcasecmp (optarg, "blackonsilver") ||
		   !strcasecmp (optarg, "black-silver") ||
		   !strcasecmp (optarg, "blacksilver"))
	    inktype = 0x03;
	  else if (!strcasecmp (optarg, "black on yellow") ||
		   !strcasecmp (optarg, "black-on-yellow") ||
		   !strcasecmp (optarg, "blackonyellow") ||
		   !strcasecmp (optarg, "black-yellow") ||
		   !strcasecmp (optarg, "blackyellow"))
	    inktype = 0x04;
	  else if (!strcasecmp (optarg, "black on gold") ||
		   !strcasecmp (optarg, "black-on-gold") ||
		   !strcasecmp (optarg, "blackongold") ||
		   !strcasecmp (optarg, "black-gold") ||
		   !strcasecmp (optarg, "blackgold"))
	    inktype = 0x05;
	  else if (!strcasecmp (optarg, "black on green") ||
		   !strcasecmp (optarg, "black-on-green") ||
		   !strcasecmp (optarg, "blackongreen") ||
		   !strcasecmp (optarg, "black-green") ||
		   !strcasecmp (optarg, "blackgreen"))
	    inktype = 0x06;
	  else if (!strcasecmp (optarg, "black on fluogreen") ||
		   !strcasecmp (optarg, "black-on-fluogreen") ||
		   !strcasecmp (optarg, "blackonfluogreen") ||
		   !strcasecmp (optarg, "black-fluogreen") ||
		   !strcasecmp (optarg, "blackfluogreen") ||
		   !strcasecmp (optarg, "black on fluorescent green") ||
		   !strcasecmp (optarg, "black-on-fluorescent-green") ||
		   !strcasecmp (optarg, "blackonfluorescentgreen") ||
		   !strcasecmp (optarg, "black-fluorescent-green") ||
		   !strcasecmp (optarg, "blackfluorescentgreen"))
	    inktype = 0x07;
	  else if (!strcasecmp (optarg, "black on fluored") ||
		   !strcasecmp (optarg, "black-on-fluored") ||
		   !strcasecmp (optarg, "blackonfluored") ||
		   !strcasecmp (optarg, "black-fluored") ||
		   !strcasecmp (optarg, "blackfluored") ||
		   !strcasecmp (optarg, "black on fluorescent red") ||
		   !strcasecmp (optarg, "black-on-fluorescent-red") ||
		   !strcasecmp (optarg, "blackonfluorescentred") ||
		   !strcasecmp (optarg, "black-fluorescent-red") ||
		   !strcasecmp (optarg, "blackfluorescentred"))
	    inktype = 0x08;
	  else if (!strcasecmp (optarg, "white on clear") ||
		   !strcasecmp (optarg, "white-on-clear") ||
		   !strcasecmp (optarg, "whiteonclear") ||
		   !strcasecmp (optarg, "white-clear") ||
		   !strcasecmp (optarg, "whiteclear"))
	    inktype = 0x09;
	  else if (!strcasecmp (optarg, "white on black") ||
		   !strcasecmp (optarg, "white-on-black") ||
		   !strcasecmp (optarg, "whiteonblack") ||
		   !strcasecmp (optarg, "white-black") ||
		   !strcasecmp (optarg, "whiteblack"))
	    inktype = 0x0a;
	  else if (!strcasecmp (optarg, "blue on white") ||
		   !strcasecmp (optarg, "blue-on-white") ||
		   !strcasecmp (optarg, "blueonwhite") ||
		   !strcasecmp (optarg, "blue-white") ||
		   !strcasecmp (optarg, "bluewhite") ||
		   !strcasecmp (optarg, "blue on clear") ||
		   !strcasecmp (optarg, "blue-on-clear") ||
		   !strcasecmp (optarg, "blueonclear") ||
		   !strcasecmp (optarg, "blue-clear") ||
		   !strcasecmp (optarg, "blueclear"))
	    inktype = 0x0b;
	  else if (!strcasecmp (optarg, "red on white") ||
		   !strcasecmp (optarg, "red-on-white") ||
		   !strcasecmp (optarg, "redonwhite") ||
		   !strcasecmp (optarg, "red-white") ||
		   !strcasecmp (optarg, "redwhite") ||
		   !strcasecmp (optarg, "red on clear") ||
		   !strcasecmp (optarg, "red-on-clear") ||
		   !strcasecmp (optarg, "redonclear") ||
		   !strcasecmp (optarg, "red-clear") ||
		   !strcasecmp (optarg, "redclear"))
	    inktype = 0x0c;

	  else if (strspn (optarg, "0123456789") != strlen (optarg))
	    {
	      fprintf (stderr,
		       "%s option error: Inktype %s unknown\n",
		       PROGNAME, optarg);
	      exit (1);
	    }
	  else
	    {
	      inktype = atoi (optarg);
	      if (inktype < 0 || inktype > 255)
		{
		  fprintf (stderr,
			   "%s option error: Inktype %s unsupported\n",
			   PROGNAME, optarg);
		  exit (1);
		}
	    }
	  break;

	case 'l':
	  labelwb = atoi (optarg);
	  if (labelwb <= 0 || labelwb > BUFLEN || labelwb > 255)
	    /* 255: we only have 1 byte for lengths! */
	    {
	      fprintf (stderr,
		       "%s option error: Labelwidth %d unsupported\n",
		       PROGNAME, labelwb);
	      exit (1);
	    }
	  break;

	case 'p':
	  if (!strcasecmp (optarg, "6mm") || !strcasecmp (optarg, "1/4in"))
	    papersize = 0;
	  else if (!strcasecmp (optarg, "9mm") ||
		   !strcasecmp (optarg, "3/8in"))
	    papersize = 1;
	  else if (!strcasecmp (optarg, "12mm") ||
		   !strcasecmp (optarg, "1/2in"))
	    papersize = 2;
	  else if (!strcasecmp (optarg, "18mm") ||
		   !strcasecmp (optarg, "19mm") ||
		   !strcasecmp (optarg, "3/4in"))
	    papersize = 3;
	  else if (!strcasecmp (optarg, "24mm") ||
		   !strcasecmp (optarg, "1in"))
	    papersize = 4;

	  else if (strspn (optarg, "0123456789") != strlen (optarg))
	    {
	      fprintf (stderr,
		       "%s option error: Papersize %s unknown\n",
		       PROGNAME, optarg);
	      exit (1);
	    }
	  else
	    {
	      papersize = atoi (optarg);
	      if (papersize <= 0 || papersize > 255)
		{
		  fprintf (stderr,
			   "%s option error: Papersize %s unsupported\n",
			   PROGNAME, optarg);
		  exit (1);
		}
	    }
	  break;

	case 't':
	  passthrough = 1;
	  break;

	case 'T':
	  passthrough = 2;
	  break;


	default:
	  usage (1);
	}
    }


  /* read/parse header */

  while (1)
    {
      cp = fgets0len (buf, BUFLEN, stdin, &i);
      if (cp == NULL)
	/* read error or EOF during header */
	return 0;

      if (!p4seen)
	{
	  if (strcmp (buf, "P4\n"))
	    {
	      if (passthrough == 0)
		{
		  fprintf (stderr,
			   "%s error: Input not in binary PBM format (P4), aborting  (use -t/-T for passthrough)\n",
			   PROGNAME);
		  exit (1);
		}

	      /* In printfile-only passthrough, check if this is a
	         reasonable printfile */
	      if (passthrough == 1)
		{
		  /* can have "00 1b 43 xx 1b" or "xx 1b 43 00 1b" or both */
		  if (i < 5)
		    {
		      int j;
		      j = fread (bufu + i, 1, BUFLEN - i, stdin);
		      if (j > 0)
			i += j;
		    }

		  /* error on short file, or non-"xx 1b 43 xx 1b" */
		  if (i < 5
		      || !(bufu[1] == 0x1b && bufu[2] == 0x43
			   && bufu[4] == 0x1b))
		    {
		      fprintf (stderr,
			       "%s error: Input not in binary PBM format (P4) or printfile, aborting  (use -T for full passthrough)\n",
			       PROGNAME);
		      exit (1);
		    }
		}

	      /* full passthrough now */
	      /* first output the remaining buffer */
	      fwrite (bufu, 1, i, stdout);	/* don't bother about errors */

	      /* then the rest */
	      while (!feof (stdin))
		{
		  i = fread (bufu, 1, BUFLEN, stdin);
		  if (i > 0)
		    fwrite (bufu, 1, i, stdout);
		  else if (feof (stdin))
		    return 0;
		  else
		    return 1;	/* file error */
		}
	      return 0;
	    }
	  p4seen = 1;
	  continue;
	}


      /* can't check this before P4 test, because we'd then lose
         data in pass-through mode.. */
      if (buf[0] == '#' || buf[0] == '\n' || buf[0] == '\0')
	/* comment, empty line, no data(?) */
	continue;

      if (p4seen)
	/* line should be "width lines" */
	{
	  i = sscanf (buf, "%d%d", &headerwidth, &headerlines);

	  if (i != 2 || headerwidth <= 0 || headerlines <= 0)
	    {
	      fprintf (stderr,
		       "%s error: Header error in input PBM, aborting\n",
		       PROGNAME);
	      exit (1);
	    }

	  /* finished with header */
	  break;
	}
    }

  filewb = (headerwidth + 7) >> 3;
  /* printf ("header %d (%d) %d\n", headerwidth, filewb, headerlines); */

  if (filewb > labelwb)
    {
      if (cropwidth)
	fprintf (stderr,
		 "%s notice: Cropping width from %d to %d bytes, continuing\n",
		 PROGNAME, filewb, labelwb);
      else
	{
	  fprintf (stderr,
		   "%s error: Expecting width %d, not %d bytes (did you forget to rotate?), aborting\n",
		   PROGNAME, labelwb, filewb);
	  exit (1);
	}
    }

  if (filewb < labelwb)
    labelwb = filewb;		/* we always use buf[] up to labelwb-1 */

  if (filewb > BUFLEN)
    {
      fprintf (stderr,
	       "%s error: File width %d larger than our buffer %d (recompile to enlarge), aborting\n",
	       PROGNAME, filewb, BUFLEN);
      exit (1);
    }


  /* Everything seems fine. Output header. */

  printf ("%c%c%c%c%c%c%c", papersize, 0x1b, 0x43, inktype, 0x1b, 0x42,
	  currmargin);


  /* And process input */

  while (!feof (stdin))
    {
      i = fread (bufu, 1, filewb, stdin);

      if (i != filewb)
	{
	  if (line < 0)
	    break;
	  /* else */
	  fprintf (stderr,
		   "%s error: Unexpected end of input after %d of %d lines, aborting\n",
		   PROGNAME, line, headerlines);
	  exit (1);
	}


      for (leftzero = 0; leftzero < labelwb && bufu[leftzero] == 0;
	   leftzero++);

      if (leftzero == labelwb)
	{
	  /* empty line */

	  if (doingempty >= maxempty)
	    doingempty = 0;

	  if (doingempty <= 0)
	    {
	      printf ("%c%c%c%c", 0x1b, 0x44, 0x00, 0x16);
	      doingempty = 1;
	    }
	  else
	    {
	      printf ("%c", 0x16);
	      doingempty++;
	    }
	}
      else
	{
	  /* non-empty line */

	  doingempty = 0;

	  if (currmargin != leftzero)
	    {
	      currmargin = leftzero;
	      printf ("%c%c%c", 0x1b, 0x42, currmargin);
	    }


	  for (rightzero = labelwb - 1; rightzero > 0 && bufu[rightzero] == 0;
	       rightzero--);

	  /* example:
	     .  0  1  2  3  4  5  6  7  8  9   <-byte#
	     .  00 00 00 12 34 56 78 00 00 00
	     .           |        |
	     .           |        rightzero=6
	     .           leftzero=3
	   */

	  printf ("%c%c%c%c", 0x1b, 0x44, rightzero - leftzero + 1, 0x16);
	  fwrite (bufu + leftzero, 1, rightzero - leftzero + 1, stdout);
	}

      if (line >= 0)
	{
	  line++;
	  if (line >= headerlines)
	    break;
	}
    }


  return 0;
}
